//============================================================================
// Name        : ehVoiceActivityDetection.cpp
// Author      : Eui-Hyun Kim (uihyun@gmail.com)
// Version     : 2.12.4
// Copyright   : Okuno Laboratory, Kyoto University
// Description :
//============================================================================

#include <iostream>
#include <string>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <libharkio/libharkio.h>

using namespace std;
using namespace FD;

class VoiceActivityDetection;

DECLARE_NODE(VoiceActivityDetection);
/*Node
 *
 * @name VoiceActivityDetection
 * @category HARK:BINAURAL+
 * @description Statistical Model-Based Voice Activity Detection
 *
 * @input_name AUDIO_SPECTRUM
 * @input_type Matrix<complex<float> >
 * @input_description Windowed spectrum data. A row index is channel, and a column index is frequency.
 *
 * @output_name VAD_DECISION
 * @output_type Vector<ObjectRef>
 * @output_description Decision of speech-present frame.
 *
 * @parameter_name VAD_NOISE_DURATION
 * @parameter_type float
 * @parameter_value 3.0
 * @parameter_description Time duration to be regarded as "noise" from the first frame [default: 3.0 seconds].
 *
 * @parameter_name VAD_THRESHOLD
 * @parameter_type float
 * @parameter_value 50.0
 * @parameter_description Threshold for voice activity decision [default: 50.0].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The length in sample between a frame and a previous frame [default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * END*/

class VoiceActivityDetection : public BufferedNode {

  int spectrumInputID;
  int vadDecisionOutputID;
  int advance;
  int samplingRate;

  // voice activity detection
  float* VadVarNoisePowerSpectrum;
  float* VadEvgNoisePowerSpectrum;
  float** VadBufferNoisePowerSpectrum;
  int* vadDecisionBuffer;
  float vadNoiseTime;
  int VadNumberOfNoiseFrames;
  float VadThreshold;

public:
  VoiceActivityDetection(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    spectrumInputID = addInput("AUDIO_SPECTRUM");
    vadDecisionOutputID = addOutput("VAD_DECISION");
    advance = dereference_cast<int>(parameters.get("ADVANCE"));
    samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));

    inOrder = true;

    // voice activity detection
    vadNoiseTime = dereference_cast<float>(parameters.get("VAD_NOISE_DURATION"));
    VadNumberOfNoiseFrames = int(vadNoiseTime/float(advance)*float(samplingRate));
    VadThreshold = dereference_cast<float>(parameters.get("VAD_THRESHOLD"));
    vadDecisionBuffer = NULL;
    VadEvgNoisePowerSpectrum = NULL;
    VadVarNoisePowerSpectrum = NULL;
    VadBufferNoisePowerSpectrum = NULL;
    if (VadNumberOfNoiseFrames <= 0) VadNumberOfNoiseFrames = 1;

  }

  virtual ~VoiceActivityDetection()
  {
    delete [] vadDecisionBuffer;
    delete [] VadVarNoisePowerSpectrum;
    delete [] VadEvgNoisePowerSpectrum;
    for (int frame = 0; frame < VadNumberOfNoiseFrames; frame++)
      delete [] VadBufferNoisePowerSpectrum[frame];
    delete [] VadBufferNoisePowerSpectrum;
  }

  void calculate(int output_id, int count, Buffer &out)
  {

    /////////////////////////////////////////////////
    // Set input spectrum data
    RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(spectrumInputID, count);
    Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;
    int nCols = spectrumInput.ncols()-1;
    const int tarMic = 1;
    const int refMic = 0;


    // ----------  Voice Activity Detection : Begin ----------

    // The telephone frequency for human voice ranges from about 300 to 3400 Hz
    int frequencyBinFrom  = 300 / (samplingRate/2/nCols);
    int frequencyBinTo = 3400 / (samplingRate/2/nCols);
    int nBinsForAveraging = frequencyBinTo - frequencyBinFrom;

    // Initialization
    float logLikelihoodRatio = 0.0;
    if (count == 0)
      {
	VadEvgNoisePowerSpectrum = new float[nCols];
	for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	  VadEvgNoisePowerSpectrum[iBin] = 0.f;

	VadVarNoisePowerSpectrum = new float[nCols];
	for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	  VadVarNoisePowerSpectrum[iBin] = 0.f;

	VadBufferNoisePowerSpectrum = new float*[VadNumberOfNoiseFrames];
	for (int frame = 0; frame < VadNumberOfNoiseFrames; frame++)
	  VadBufferNoisePowerSpectrum[frame] = new float[nCols];

      } else if (count < VadNumberOfNoiseFrames) {

      ////////////////////////////////////////////////
      // Noise spectrum extraction
      for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	{
	  VadBufferNoisePowerSpectrum[count][iBin] = (spectrumInput(tarMic,iBin).real()*spectrumInput(tarMic,iBin).real() + spectrumInput(tarMic,iBin).imag()*spectrumInput(tarMic,iBin).imag()) / 2;
	  VadBufferNoisePowerSpectrum[count][iBin] += (spectrumInput(refMic,iBin).real()*spectrumInput(refMic,iBin).real() + spectrumInput(refMic,iBin).imag()*spectrumInput(refMic,iBin).imag()) / 2;

	  VadEvgNoisePowerSpectrum[iBin] += VadBufferNoisePowerSpectrum[count][iBin] / VadNumberOfNoiseFrames;
	}

    } else if (count == VadNumberOfNoiseFrames) {

      for (int frame = 0; frame < VadNumberOfNoiseFrames; frame++)
	for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	  VadVarNoisePowerSpectrum[iBin] += abs(VadBufferNoisePowerSpectrum[frame][iBin] - VadEvgNoisePowerSpectrum[iBin]) / VadNumberOfNoiseFrames;

    } else {


      ////////////////////////////////////////////////
      // Statistical Model-Based Voice Activity Detection
      float noisyPowerSpectrum = 0.f;
      float SNRpost = 0.f;

      for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	{
	  noisyPowerSpectrum = (spectrumInput(tarMic,iBin).real()*spectrumInput(tarMic,iBin).real() + spectrumInput(tarMic,iBin).imag()*spectrumInput(tarMic,iBin).imag()) / 2;
	  noisyPowerSpectrum += (spectrumInput(refMic,iBin).real()*spectrumInput(refMic,iBin).real() + spectrumInput(refMic,iBin).imag()*spectrumInput(refMic,iBin).imag()) / 2;

	  SNRpost = noisyPowerSpectrum / VadVarNoisePowerSpectrum[iBin];

	  if (SNRpost < 0)
	    SNRpost = 0.f;

	  logLikelihoodRatio += (SNRpost - log(SNRpost) - 1) / nBinsForAveraging;
	}

    }

    // VAD decision on a current frame
    bool vadDecision = false; // noise
    if (count >= VadNumberOfNoiseFrames && logLikelihoodRatio > VadThreshold)
      vadDecision = true; // speech

    // ----------  Voice Activity Detection : End ----------


    /////////////////////////////////////////////////
    // Set output data
    RCPtr<Vector<ObjectRef> > vadDecisionPtr(new Vector<ObjectRef>);
    (*outputs[vadDecisionOutputID].buffer)[count] = vadDecisionPtr;

    if (vadDecision) {
      RCPtr<Source> vadSrcOutput(new Source());
      vadSrcOutput->power = 1.f; // Power of the direction localized
      vadSrcOutput->remaining =  0.f; // An index that assumes how long the localized sound source lasts from then on
      vadSrcOutput->id = 777; // id #777 indicates decisions of voice activity detection.
      float theta = 0 * M_PI / 180.0f;  // angle
      float phi = 0 * M_PI / 180.0f;   // elevation
      float cos_phi = cos(phi);
      vadSrcOutput->x[0] = (float)(cos(theta) * cos_phi);
      vadSrcOutput->x[1] = (float)(sin(theta) * cos_phi);
      vadSrcOutput->x[2] = (float)(sin(phi));
      vadDecisionPtr->push_back(vadSrcOutput);
    }

  }
};

